/*
 * Copyright 2024 KylinSoft Co., Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#include "xunfeinlpengine_p.h"
#include <unistd.h>
#include <curl/websockets.h>
#include "servererror.h"
#include "token.h"
#include "util.h"

bool XunfeiNlpEnginePrivate::connectWebsocket(std::string hostUrl) {
    std::string callurl =
        xunfei_nlp_token::getAuthenticationUrl(apiKey_, secretKey_, hostUrl);

    CURLcode res;
    if (currentCurl_) {
        curl_easy_setopt(currentCurl_, CURLOPT_URL, callurl.c_str());
        curl_easy_setopt(currentCurl_, CURLOPT_CONNECT_ONLY, 2L);
        res = curl_easy_perform(currentCurl_);
        if (res == CURLE_HTTP_RETURNED_ERROR) {
            fprintf(stderr, "xunfei authenticate failed: %s\n",
                    curl_easy_strerror(res));
            currentError_ = ai_engine::lm::EngineError(
                ai_engine::lm::AiCapability::Nlp,
                ai_engine::lm::EngineErrorCategory::Initialization,
                (int)ai_engine::lm::NlpEngineErrorCode::Unauthorized,
                "接口鉴权失败");
            return false;
        }
        if (res != CURLE_OK) {
            fprintf(stderr, "net error: %s\n", curl_easy_strerror(res));
            currentError_ = ai_engine::lm::EngineError(
                ai_engine::lm::AiCapability::Nlp,
                ai_engine::lm::EngineErrorCategory::Initialization,
                (int)ai_engine::lm::NlpEngineErrorCode::NetworkDisconnected,
                curl_easy_strerror(res));
            return false;
        }
        return true;
    }
    currentError_ = ai_engine::lm::EngineError(
        ai_engine::lm::AiCapability::Nlp,
        ai_engine::lm::EngineErrorCategory::Initialization,
        (int)ai_engine::lm::NlpEngineErrorCode::InvalidWords, "创建资源失败");
    return false;
}

void onWebSocket(bool (*doReceiveData)(XunfeiNlpEnginePrivate *self),
                 XunfeiNlpEnginePrivate *self) {
    while (true) {
        if (doReceiveData(self)) {
            usleep(1000);
            continue;
        }
        return;
    }
}

void XunfeiNlpEnginePrivate::closeWebsocket() {
    size_t sent;
    if (currentCurl_) {
        curl_ws_send(currentCurl_, "", 0, &sent, 0, CURLWS_CLOSE);
    }
}

const std::string XunfeiNlpEnginePrivate::systemRole_{
    "我是您的 AI 助手。"
    "您可以通过文字和语音对话与我进行互动。我具备跨应用文字处理和系统控制功能，"
    "无论是工作学习还是日常生活，我都能为您提供帮助。"};

XunfeiNlpEnginePrivate::XunfeiNlpEnginePrivate() {}

XunfeiNlpEnginePrivate::~XunfeiNlpEnginePrivate() {
    clearContext();
    streamChatData_.clear();
}

std::string XunfeiNlpEnginePrivate::engineName() { return "xunfei"; }

void XunfeiNlpEnginePrivate::setConfig(const std::string &config) {
    Json::Value configJson = xunfei_nlp_util::formatJsonFromString(config);
    if (configJson.isNull() || !configJson.isMember("apiKey") ||
        !configJson.isMember("secretKey") || !configJson.isMember("appId") ||
        !configJson["apiKey"].isString() ||
        !configJson["secretKey"].isString() ||
        !configJson["appId"].isString()) {
        std::fprintf(stderr, "Invalid config for xunfei chat engine: %s\n",
                     config.c_str());
        return;
    }
    appId_ = configJson["appId"].asString();
    apiKey_ = configJson["apiKey"].asString();
    secretKey_ = configJson["secretKey"].asString();

    context_["header"]["app_id"] = appId_;
}

void XunfeiNlpEnginePrivate::setChatResultCallback(
    ai_engine::lm::nlp::ChatResultCallback callback) {
    chatResultCallback_ = callback;
}

void XunfeiNlpEnginePrivate::setContextSize(int size) { contextSize_ = size; }

void XunfeiNlpEnginePrivate::clearContext() {
    context_["payload"]["message"]["text"].clear();
    addContext_("system", systemRole_);
}

bool XunfeiNlpEnginePrivate::doSendData(const char *data, int dataLength,
                                        int type) {
    size_t sent;
    CURLcode resultCode =
        curl_ws_send(currentCurl_, data, dataLength, &sent, 0, type);
    if (resultCode != CURLE_OK) {
        fprintf(stderr, "send data to xunfei failed: %s\n",
                curl_easy_strerror(resultCode));
        currentError_ = ai_engine::lm::EngineError(
            ai_engine::lm::AiCapability::Nlp,
            ai_engine::lm::EngineErrorCategory::Initialization,
            (int)ai_engine::lm::NlpEngineErrorCode::NetworkDisconnected,
            curl_easy_strerror(resultCode));
        return false;
    }
    return true;
}

bool XunfeiNlpEnginePrivate::doReceiveChatData() {
    size_t rlen;
    const curl_ws_frame *meta;
    char buffer[4100]{0};

    if (isStopped_) {
        return false;
    }

    CURLcode curlCode =
        curl_ws_recv(currentCurl_, buffer, sizeof(buffer), &rlen, &meta);
    if (curlCode == CURLE_AGAIN || curlCode == CURLE_RECV_ERROR) {
        return true;
    }
    if (curlCode != CURLE_OK) {
        fprintf(stderr, "net error: %s\n", curl_easy_strerror(curlCode));
        currentError_ = ai_engine::lm::EngineError(
            ai_engine::lm::AiCapability::Nlp,
            ai_engine::lm::EngineErrorCategory::Initialization,
            (int)ai_engine::lm::NlpEngineErrorCode::NetworkDisconnected,
            curl_easy_strerror(curlCode));
        ai_engine::lm::nlp::ChatResult result{std::string{}, currentError_};
        chatResultCallback_(result);
        return false;
    }

    if (int errorCode =
            xunfei_nlp_server_error::parseErrorCode(std::string(buffer))) {
        fprintf(stderr, "xunfei chat failed %s\n", std::string(buffer).c_str());
        auto errorTuple =
            xunfei_nlp_server_error::errorCode2nlpResult(errorCode);
        currentError_ = ai_engine::lm::EngineError(
            ai_engine::lm::AiCapability::Nlp, std::get<0>(errorTuple),
            (int)std::get<1>(errorTuple), std::string(buffer));
        ai_engine::lm::nlp::ChatResult result{std::string{}, currentError_};
        chatResultCallback_(result);
        return false;
    }

    Json::Value obj =
        xunfei_nlp_util::formatJsonFromString(std::string(buffer));

    Json::Value callbackObj;
    callbackObj["sentence_id"] = obj["payload"]["choices"]["seq"];
    callbackObj["is_end"] = (obj["header"]["status"].asInt() == 2);
    callbackObj["result"] = obj["payload"]["choices"]["text"][0]["content"];
    streamChatData_.push_back(obj["payload"]["choices"]["text"][0]["content"]);
    ai_engine::lm::nlp::ChatResult result{callbackObj.toStyledString(),
                                          ai_engine::lm::EngineError()};
    chatResultCallback_(result);
    if (obj["header"]["status"].asInt() == 2) {
        return false;
    }

    return true;
}

void XunfeiNlpEnginePrivate::addContext_(const std::string &role,
                                         const std::string &message) {
    int size = context_["payload"]["message"]["text"].size();
    context_["payload"]["message"]["text"][size]["role"] = role;
    context_["payload"]["message"]["text"][size]["content"] = message;
}

void XunfeiNlpEnginePrivate::removeLastContext_() {
    int size = context_["payload"]["message"]["text"].size();
    if (size - 1 >= 0 &&
        context_["payload"]["message"]["text"][size - 1]["role"].asString() ==
            "user") {
        Json::Value removeValue;
        context_["payload"]["message"]["text"].removeIndex(size - 1,
                                                           &removeValue);
        return;
    }
    if ((size - 1) >= 0 && (size - 2) >= 0 &&
        context_["payload"]["message"]["text"][size - 1]["role"].asString() ==
            "assistant" &&
        context_["payload"]["message"]["text"][size - 2]["role"].asString() ==
            "user") {
        Json::Value removeValue;
        context_["payload"]["message"]["text"].removeIndex(size - 1,
                                                           &removeValue);
        context_["payload"]["message"]["text"].removeIndex(size - 2,
                                                           &removeValue);
        return;
    }
}

bool XunfeiNlpEnginePrivate::initChatModule(ai_engine::lm::EngineError &error) {
    currentError_ = error;
    if (!initResources() || !connectWebsocket(hostUrl_)) {
        error = currentError_;
        return false;
    }
    closeWebsocket();
    releaseResources();
    context_["parameter"]["chat"]["domain"] = "generalv3.5";
    addContext_("system", systemRole_);
    return true;
}

bool XunfeiNlpEnginePrivate::chat(const std::string &message,
                                  ai_engine::lm::EngineError &error) {
    isStopped_ = false;
    currentError_ = error;

    if (!chatResultCallback_) {
        return false;
    }
    if (message.empty()) {
        error = ai_engine::lm::EngineError(
            ai_engine::lm::AiCapability::Nlp,
            ai_engine::lm::EngineErrorCategory::Initialization,
            (int)ai_engine::lm::NlpEngineErrorCode::TooLittleData, "文本为空");
        ai_engine::lm::nlp::ChatResult result{std::string{}, currentError_};
        chatResultCallback_(result);
        return false;
    }

    if (!initResources() || !connectWebsocket(hostUrl_)) {
        error = currentError_;
        ai_engine::lm::nlp::ChatResult result{std::string{}, currentError_};
        chatResultCallback_(result);
        return false;
    }

    if (contextSize_ == 0) {
        clearContext();
    } else if (contextSize_ > 0 && contextSize_ <= 8) {
        context_["parameter"]["chat"]["max_tokens"] = contextSize_ * 1024;
    } else {
        context_["parameter"]["chat"]["max_tokens"] = 2048;
    }

    fprintf(stdout, "%s\n", context_.toStyledString().c_str());

    addContext_("user", message);
    std::string postString = context_.toStyledString();
    if (!doSendData(postString.c_str(), postString.size(), CURLWS_TEXT)) {
        error = currentError_;
        removeLastContext_();
        streamChatData_.clear();
        closeWebsocket();
        ai_engine::lm::nlp::ChatResult result{std::string{}, currentError_};
        chatResultCallback_(result);
        return false;
    }

    while (doReceiveChatData()) {
        usleep(1000);
        continue;
    }
    if (!currentStatus()) {
        removeLastContext_();
        streamChatData_.clear();
        if (!isStopped_) {
            ai_engine::lm::nlp::ChatResult result{std::string{}, currentError_};
            chatResultCallback_(result);
        }
    } else {
        std::string assistantcontent;
        for (Json::Value &value : streamChatData_) {
            assistantcontent.append(value.asString());
        }
        addContext_("assistant", assistantcontent);
        streamChatData_.clear();
    }
    closeWebsocket();
    error = currentError_;
    return currentStatus();
}

void XunfeiNlpEnginePrivate::stopChat() { isStopped_ = true; }

bool XunfeiNlpEnginePrivate::destroyChatModule(
    ai_engine::lm::EngineError &error) {
    stopChat();
    return true;
}

bool XunfeiNlpEnginePrivate::initResources() {
    if (currentCurl_) {
        curl_easy_cleanup(currentCurl_);
        currentCurl_ = nullptr;
    }
    currentCurl_ = curl_easy_init();

    return true;
}

void XunfeiNlpEnginePrivate::releaseResources() {
    isStopped_ = true;
    if (currentCurl_) {
        curl_easy_cleanup(currentCurl_);
        currentCurl_ = nullptr;
    }
}

bool XunfeiNlpEnginePrivate::currentStatus() {
    return (currentError_.getCode() == -1);
}
